    # Source for the CSS/HTML part:
# - http://www.nicolashoening.de/?twocents&nr=8
# - http://www.standards-schmandards.com/exhibits/barchart/
# - http://pythoughts.com/table-style-css/

try:
    from xml_serializer.xml_serializer import XMLSerializer
except ImportError:
    pass
    
from collections import defaultdict
from operator import itemgetter

html = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML '\
    '1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n'

popitup = """function popitup(id_) {
  newwindow2=window.open('','name','height=400,width=30d0');
  var elem = document.getElementById(id_);
  var newElem = elem.cloneNode(true);
  newElem.setAttribute("style", "");

  var tmp = document.createElement("div");
  tmp.appendChild(newElem);

  var doc = newwindow2.document;
  doc.write('<html><head><title>popup</title>');
  doc.write('<link rel="stylesheet" href="html/screen.css">');
  doc.write('</head><body>');
  doc.write(tmp.innerHTML);
  doc.write('</body></html>');
  doc.close();
}"""


def hamming_loss(example, quiet=True):
    """
    Estimates hamming loss.

    Each example is a pair of label sequences (predicted_labels, gold_labels).

    If the gold and predicted sequences are always the same, the hamming loss
    is 0; if they have never a common label, it is 1.

    This function either compute the “true” Hamming loss (that requires a gold
    sequence) or an “ambiguous” loss in which only a set of possible label is
    known at each position.
    """
    correct = 0
    total = 0

    for predicted, gold in example:
        correct += sum(1 for p, g in zip(predicted, gold) if p == g or p in g)
        total += len(predicted)

    try:
        if not quiet:
            print("Hamming: {:.2%}".format(1 - correct / total))

        return 1 - correct / total
    except ZeroDivisionError:
        return "∞"


def amb_hamming_loss(predicted, gold, possible_labels):
    """
    Compute the hamming loss only for observation that are ambiguous

    Paramaters
    ----------
    - predicted a sequence of sequences of labels
      the predicted labels for the *whole* corpus
    - gold, a sequence of sequences of labels
      the predicted labels for the *whole* corpus
    - possible_labels, a sequence of set of labels
      the i-th element of this set corresponds to the licenced labels
      of the i-th observation. Only the observations for which more
      than one label is licenced are considered.
    """

    correct = 0
    total = 0
    for pred_seq, gold_seq, poss_seq in zip(predicted, gold, possible_labels):
        filtered = [(p, g) for p, g, pp in zip(pred_seq, gold_seq, poss_seq) if len(pp) > 1]
        total += len(filtered)
        correct += sum(1 for p, g in filtered if p == g)

    return correct / total


def matrix2table(matrix):
    """
    Transform a confusion matrix into a table
    """

    labels = sorted(matrix.keys())
    table = [[{"value": "↓ pred/gold →"}] + [{"value": l} for l in labels]]
    for pred in labels:
        row = [{"value": pred}] + [{"value": matrix[pred][g]} for g in labels]
        table.append(row)

    return table


def confusion_matrix(gold_sequences, predicted_sequences):
    """
    Return a dictionnary so that `dict[l1][l2]` corresponds to the number of
    time `l1` was predicted while the correct answer was `l2`
    """

    from itertools import chain, product

    all_labels = set(chain(*(gold_sequences + predicted_sequences)))

    count = defaultdict(lambda: defaultdict(int))
    for gold_labels, pred_labels in zip(gold_sequences, predicted_sequences):
        for gold, pred in zip(gold_labels, pred_labels):
            count[pred][gold] += 1

    res = {}
    for l1, l2 in product(all_labels, repeat=2):
        if l1 not in res:
            res[l1] = {}
        res[l1][l2] = count[l1][l2]

    return res


def bar_chart(id_, val, doc):
    """
    Build a bar graph from a dictionnary

    Parameters
    ----------
    - id_, a string
    - val, a dictionnary
        maps a name to value
    - doc, a stream
        where the html code has to be written to
    """

    max_value = max(val.values())
    val = sorted(val.items(), key=itemgetter(1), reverse=True)

    doc.start_element("div", id=id_, style="display:none;")
    doc.start_element("table", {"class": "popup"})

    for name, value in val:
        doc.start_element("tr")
        doc.text_element("td", name)

        doc.start_element("td", {"class": "value"})
        doc.start_element("img", src="html/bar.png", width=str(100 * value // max_value), height="16")
        doc.end_element("img")
        doc.add_text(value)
        doc.end_element("td")
        doc.end_element("tr")

    doc.end_element("table")
    doc.end_element("div")


def html_header(doc):

    doc.start_element("head")

    doc.start_element("meta")
    doc.add_attributes(charset="utf-8")
    doc.end_element("meta")

    doc.start_element("link")
    doc.add_attributes(href="./html/screen.css", rel="stylesheet",
                       type="text/css", media="screen")
    doc.end_element("link")

    doc.start_element("script")
    doc.add_attributes(type="text/javascript", src="./html/jquery_1.10.2.js")
    doc.add_text(" ")
    doc.end_element("script")

    doc.start_element("script")
    doc.add_attributes(type="text/javascript", src="./html/nhpup_1.1.js")
    doc.end_element("script")

    doc.start_element("script")
    doc.add_attributes(type="text/javascript")
    doc.add_text(popitup)
    doc.end_element("script")

    doc.end_element("head")


def f1_table(data, doc):

    doc.start_element("div")

    doc.text_element("h1", "Confusion Matrix")

    gold = [d["gold"] for d in data]
    pred = [d["pred"] for d in data]

    table = matrix2table(confusion_matrix(gold, pred))
    table = make_table(table, doc)

    doc.end_element("div")


def metrics(data, doc):

    gold = [d["gold"] for d in data]
    pred = [d["pred"] for d in data]
    amb = [d["ambiguous_label"] for d in data]

    h = hamming_loss(zip(pred, gold))
    ha = hamming_loss(zip(gold, amb))

    doc.start_element("div")

    doc.start_element("h1")
    doc.add_text("Evaluation")
    doc.end_element("h1")

    doc.start_element("ul")
    doc.start_element("li")
    doc.add_text("Hamming: {:.2%}".format(h))
    doc.end_element("li")
    doc.start_element("li")
    doc.add_text("% of gold labels not in constraints: {:.2%}".format(ha))
    doc.end_element("li")
    doc.end_element("ul")

    doc.end_element("div")


def make_report(data, modules, output):

    with XMLSerializer(output, dtd=html) as doc:
        doc.start_element("html")

        html_header(doc)

        doc.start_element("body")
        for module in modules:
            print(module)
            m = eval(module)
            m(data, doc)
        doc.end_element("body")

        doc.end_element("html")


def dump_ambiguous_labels(data, doc):
    from sequence_labeling.pos_tagger import normalize

    for sent_id, d in enumerate(data):
        words = d["words"]
        norm_words = [normalize(w, None) for w in words]
        amb_labels = d["ambiguous_label"]

        for w, n, a in zip(words, norm_words, amb_labels):
            print("\t".join([w, n, str(a)]))


def dump_errors(data, doc):

    from collections import Counter
    c = Counter()

    for sent_id, d in enumerate(data):
        words = d["words"]
        predicted_labels = d["pred"]
        gold_labels = d["gold"]
        
        data = zip(words, predicted_labels, gold_labels)
        c.update((w, p, g) for w, p, g in data if g != p)

    from pprint import pprint
    pprint(c.most_common(300))



def sentences(data, doc):

    doc.start_element("div")
    doc.text_element("h1", "Error details")

    for sent_id, d in enumerate(data):

        words = [{"value": "words", "style": "font-weight:bold;"}]
        count = 0
        for w, e, f in zip(d["words"], d["errors"], d["features"]):

            id_ = "feat{}-{}".format(sent_id, count)
            bar_chart(id_, f, doc)
            cmd = "nhpup.popup($('#{}').html());".format(id_)

            count += 1

            cell = {"value": w,
                    "onclick": "return popitup('{}');".format(id_),
                    "onmouseover": cmd}
            if e:
                cell["class"] = "error"

            words.append(cell)

        gold = [{"value": "gold", "style": "font-weight:bold;"}]
        gold.extend([{"value": w} for w in d["gold"]])

        pred = [{"value": "pred", "style": "font-weight:bold;"}]
        for count, (label, s) in enumerate(zip(d["pred"], d["labels_scores"])):
            id_ = "label{}-{}".format(sent_id, count)
            bar_chart(id_, s, doc)
            cmd = "nhpup.popup($('#{}').html());".format(id_)
            pred.append({"value": label,
                        "onclick": "return popitup('{}');".format(id_),
                        "onmouseover": cmd})

        amb_labels = [{"value": "supervision", "style": "font-weight:bold;"}]
        for dd, e in zip(d["ambiguous_label"], d["amb_errors"]):
            cell = {"value": dd}

            if e:
                cell["class"] = "error"

            amb_labels.append(cell)

        table = [words, gold, pred, amb_labels]

        make_table(table, doc)

    doc.end_element("div")


def make_table(data, doc):
    """
    data is a list of rows. Each row is a list of cell.
    """
    doc.start_element("table")
    for seq in data:

        doc.start_element("tr")

        for s in seq:
            doc.start_element("td")

            attrib = (a for a in s if a != "value")
            attrib = {name: s[name] for name in attrib}

            doc.add_attributes(**attrib)
            doc.add_text(str(s["value"]))
            doc.end_element("td")

        doc.end_element("tr")

    doc.end_element("table")
