import random
import pickle

from itertools import zip_longest
from collections import defaultdict, deque


class AveragedPerceptron:

    def __init__(self):

        # self.weights[class][features] --> importance of `feature` to
        # predict `class`
        self.weights = defaultdict(lambda: defaultdict(float))
        # the accumulated value for a pair class/feature
        self._accum = defaultdict(lambda: defaultdict(float))
        # the last time a pair class/feature was updated
        self._last_update = defaultdict(lambda: defaultdict(int))
        # number of instances seen
        self.n_updates = 0

    def __perceptron_update(self, f, l, how_much):
        self._accum[l][f] += (self.n_updates -
                              self._last_update[l][f] - 1) * self.weights[l][f]
        self._last_update[l][f] = self.n_updates
        self.weights[l][f] += how_much
        self._accum[l][f] += self.weights[l][f]

    def update(self, truth, guess, features):

        self.n_updates += 1

        if truth == guess:
            return

        for f in features:
            self.__perceptron_update(f, truth, +1)
            self.__perceptron_update(f, guess, -1)

    def labels_scores(self, features):

        scores = defaultdict(float)
        for c in self.weights:
            scores[c] = sum(features[f] * self.weights[c][f] for f in features)

        return scores

    def predict(self, features, possible_labels=None):
        """
        Predict the label associated to a feature vector

        Parameters
        ----------
        - features, a dictionnary
            a dictionnary mapping feature names to their value
        - possible_labels, an iterable
            if not None, only the set of possible labels (i.e. the argmax of
            the decision rule) is reduced to this set
        """
        scores = self.labels_scores(features)
        if possible_labels is not None:
            scores = {k: v for k, v in scores.items() if k in possible_labels}

        # secondary alphabetic sort for stability
        try:
            return max(scores, key=lambda label: (scores[label], label))
        except ValueError:
            # XXX This is happening at the very first update when training a
            # a POSTagger: as the weights vector is empty, we do not know any
            #  label that can be returned
            return "NOUN"

    def average_weights(self):

        new_weigths = defaultdict(lambda: defaultdict(float))
        for label in self.weights:
            for f in self.weights[label]:
                new_weigths[label][f] = self._accum[label][f]
                a = (self.n_updates - self._last_update[label][f]) * self.weights[label][f]
                new_weigths[label][f] += a
                new_weigths[label][f] /= self.n_updates

        self.weights = new_weigths

    def __getstate__(self):
        # should we also serialize the other attributes to allow learning to
        # continue.
        return {"weights": {k: v for k, v in self.weights.items()}}

    def __setstate__(self, dict):
        self.weights = defaultdict(lambda: defaultdict(float), dict["weights"])

    def load(self, filehandler):
        self.weights = pickle.load(filehandler)


class WeakilySupervisedPerceptron(AveragedPerceptron):

    def update(self, truth, guess, features):
        self.n_updates += 1

        if guess in truth:
            return

        for f in features:

            def u(l, how_much):
                a = (self.n_updates - self._last_update[l][f] - 1) * self.weights[l][f]
                self._accum[l][f] += a
                self._last_update[l][f] = self.n_updates
                self.weights[l][f] += how_much
                self._accum[l][f] += self.weights[l][f]

            for possible in truth:
                u(possible, +1)

            u(guess, -1)


class SkipPerceptron(AveragedPerceptron):

    def update(self, truth, guess, features):

        if len(truth) != 1:
            return

        truth = list(truth)[0]

        super().update(truth, guess, features)


class POSTagger:

    def __init__(self,
                 feat_template,
                 average=False,
                 model_factory=None,
                 model=None):
        """
        A simple POS-tagger that relies on an history-based model.

        Parameters
        -----------

        - feat_template, a callable
          a function extracting the features describing an element
        - model_factory, a callable
          a function that create a new predictor used to make the local
          decisions
        - model, an object
          an object used to make the local decisions. Passing such an object
          allows to use the tagger without training it first
        - average, a boolean
          if true, average the weights after training
        """
        self.features = feat_template
        self.average = average

        if model is not None:
            self.model = model
        elif model_factory is not None:
            self.model = None
            self.new_model = model_factory
        else:
            raise Exception

    def tag(self, words, possible_labels=None):
        """
        Parameters
        ----------

        - `words`: a list of string,
                   the tokenized sentence
        - `dictionary`: a list of sets
                        the set of possible labels for the i-th words

        Using a list to store the possible labels (instead of computing them on
        the fly) allows to reduce the number of parameters of this method and
        increase its re-usability.
        """
        if possible_labels is None:
            possible_labels = []

        if self.model is None:
            raise Exception

        history = deque()

        tags = []
        context = defaultdict(lambda: None, enumerate(words))

        for index, (word, p) in enumerate(zip_longest(words, possible_labels)):
            features = self.features(index, context, history)
            tag = self.model.predict(features, p)

            tags.append(tag)

            history.appendleft(tag)

        return tags

    def train(self, data, n_iter, iter_callback=None,
              possible_labels=None, listener=lambda x: x):
        """
        Parameters
        ----------

        - data, an iterable of tuples
            the training set; each tuple is a pair (words, labels)
        - n_iter, int
            the number of epoch (pass over the train set)
        - iter_callback, a callable
            a function to call at the end of all iteration. The function takes
            a single argument: the tagger
        - possible_labels, a callable
            a function that returns the set of licenced tag for a
            given word. If not `None`, only a tag of this set can be
            predicted during training.
        """
        self.model = self.new_model()
        listener.msg("n_iter = {}".format(n_iter))
        for iter_ in range(n_iter):

            random.shuffle(data)

            n_correct = 0
            n_total = 0

            count = 0
            for words, labels in listener.iter(data):

                history = deque()

                ctxt = defaultdict(lambda: None, enumerate(words))

                for index, (word, label) in enumerate(zip(words, labels)):
                    features = self.features(index, ctxt, history)

                    if possible_labels is not None:
                        guess = self.model.predict(features,
                                                   possible_labels(word))
                    else:
                        guess = self.model.predict(features)

                    self.model.update(label, guess, features)

                    history.appendleft(guess)

                    n_correct += guess == label or guess in label
                    n_total += 1

                count += 1

            listener.msg("Iter {}: {:.1%}".format(iter_, n_correct / n_total))

            if iter_callback is not None:
                weights = self.model.weights
                self.model.average_weights()
                iter_callback(self)
                self.model.weights = weights

        if self.average:
            listener.msg("Averaging weights")
            self.model.average_weights()

        listener.msg("#features: {:,}".format(sum(len(v) for v in self.model.weights.values())))
