from collections import defaultdict


DIGIT = "!DIGIT"
YEAR = "!YEAR"


def pos_features(index, context, history,
                 allowed_suffixes=None, allowed_words=None,
                 cls_dict=None, type_dict=None, history_size=2):
    """Extract features for POS tagging

    Parameters
    ---------
    - index, an int
        the position in the sentence of the word the features of which have
        to be extracted

    - context, a dictionary
        the context maps a position to a word. The dictionary must be
        designed so that accessing a position that does not exist
        (such as -1, -2 or indexes larger than the number of words in
        the sequence) will return :code:`None`

    - history, a list of labels
         the history (i.e. last predicted labels), history[0] is the
         last predicted label, history[1] the one before, ...

    - allowed_words, a set of strings
         If not `None` only the words in this set are considered to 
         define features.
    
    - allowed_suffixes
         If not `None` only the suffixes in this set are considered to 
         define features.

    - cls_dict, a dictionary
         maps a word to its brown class. If `None` brown features are
         not used.

    - type_dict, a callable
         returns the list of possible POS tags a word can have
         (i.e. type constraints). If `None` this kind of feature is
         not used.

     - history_size, an int
          size of the history to consider
    """

    features = defaultdict(int)
    features["biais"] = 1

    # structure features
    for his_pos in range(history_size):
        label = history[his_pos] if his_pos < len(history) else None
        features["tag@i-{}={}".format(his_pos + 1, label)] = 1

    prev_label = history[0] if his_pos < len(history) else None
    prev_prev_label = history[1]if his_pos < len(history) else None
    f = "tag@i-1={}_and_tag@i-2={}".format(prev_label,
                                           prev_prev_label)
    features[f] = 1

    current_word = normalize(context[index], allowed_words)
    features["tag@i-1={}_and_word@i={}".format(prev_label, current_word)] = 1

    # observation features
    words = [(delta, normalize(context[index + delta], allowed_words))
             for delta in range(-2, 3)]
    suffixes = [(delta, w[-2:]) for delta, w in words if w is not None]
    suffixes += [(delta, w[-3:]) for delta, w in words if w is not None]

    if allowed_suffixes is not None:
        suffixes = [(delta, s) for delta,
                    s in suffixes if s in allowed_suffixes]

    features.update({"w@i{0:+}={1}".format(delta, w): 1 for delta, w in words})
    features.update({"suffix@i{:+}={}".format(delta, s): 1 for delta, s in
                    suffixes})

    # class features
    if cls_dict is not None:
        for delta in range(-2, 3):
            if context[index + delta] in cls_dict:
                features["cls@{:+}={}".format(delta, cls_dict[context[index + delta]])] = 1

    # type information
    if type_dict is not None:
        for delta in range(-2, 3):
            word = context[index + delta]

            if word is None:
                continue

            for label in type_dict(word):
                features["word@{:+}_has_label_{}".format(delta, label)] = 1

    # Surface features
    if context[index].istitle() and index != 0:
        features["w@0=isUpperCaseAndNotFirst"] = 1
    else:
        features["w@0=isLowerCase"] = 1

    if "-" in context[index]:
        features["w@0hasHyphen"] = 1

    for i in range(10):
        if str(i) in context[index]:
            features["digitInWord"] = 1

    return features


def normalize(word, allowed_words):
    """
    String normalization

    Detect digit and year (number between 1800 and 2100)
    """
    if word is None:
        return word

    if word.isdigit():
        return DIGIT

    word = word.lower()

    if allowed_words is None:
        return word.lower()
    else:
        return word if word in allowed_words else "_NF_"
