from sequence_labeling.evaluation import confusion_matrix,\
    matrix2table, bar_chart, hamming_loss, amb_hamming_loss

from xml_serializer.xml_serializer import XMLSerializer


def test_same():
    """
    Confusion matrix of two identical sequences
    """

    seq = ["A B B A".split(),
           "A D B C".split()]

    res = confusion_matrix(seq, seq)

    expected_res = {'A': {'A': 3, 'C': 0, 'B': 0, 'D': 0},
                    'C': {'A': 0, 'C': 1, 'B': 0, 'D': 0},
                    'B': {'A': 0, 'C': 0, 'B': 3, 'D': 0},
                    'D': {'A': 0, 'C': 0, 'B': 0, 'D': 1}}

    assert res == expected_res


def test_different():
    """
    Confusion matrix of different sequences
    """

    gold_seq = ["A B A B".split(),
                "A D B C".split()]

    pred_seq = ["A A B B".split(),
                "A C D C".split()]

    res = confusion_matrix(gold_seq, pred_seq)

    expected_res = {'A': {'A': 2, 'C': 0, 'B': 1, 'D': 0},
                    'C': {'A': 0, 'C': 1, 'B': 0, 'D': 1},
                    'B': {'A': 1, 'C': 0, 'B': 1, 'D': 0},
                    'D': {'A': 0, 'C': 0, 'B': 1, 'D': 0}}

    assert res == expected_res


def test_missing_entry():
    """
    Confusion matrix when a label has never been predicted
    """

    gold_seq = ["A B A B".split(),
                "A D B C".split()]

    pred_seq = ["A A B B".split(),
                "A C B C".split()]

    res = confusion_matrix(gold_seq, pred_seq)

    expected_res = {'A': {'A': 2, 'C': 0, 'B': 1, 'D': 0},
                    'C': {'A': 0, 'C': 1, 'B': 0, 'D': 1},
                    'B': {'A': 1, 'C': 0, 'B': 2, 'D': 0},
                    'D': {'A': 0, 'C': 0, 'B': 0, 'D': 0}}

    assert res == expected_res


def test_matrix2list():
    """
    Transform a confusion matrix into a table
    """

    expected = []
    expected.append([{'value': '↓ pred/gold →'}, {'value': 'A'},
                     {'value': 'B'}, {'value': 'C'}, {'value': 'D'}])
    expected.append([{'value': 'A'}, {'value': 2}, {'value': 1},
                     {'value': 0}, {'value': 0}])
    expected.append([{'value': 'B'}, {'value': 1}, {'value': 2},
                     {'value': 0}, {'value': 0}])
    expected.append([{'value': 'C'}, {'value': 0}, {'value': 0},
                     {'value': 1}, {'value': 1}])
    expected.append([{'value': 'D'}, {'value': 0}, {'value': 0},
                     {'value': 0}, {'value': 0}])

    confusion = {'A': {'A': 2, 'C': 0, 'B': 1, 'D': 0},
                 'C': {'A': 0, 'C': 1, 'B': 0, 'D': 1},
                 'B': {'A': 1, 'C': 0, 'B': 2, 'D': 0},
                 'D': {'A': 0, 'C': 0, 'B': 0, 'D': 0}}

    assert matrix2table(confusion) == expected


def test_chart():
    """
    Build a simple chart
    """
    from io import StringIO
    output = StringIO()

    doc = XMLSerializer(output)
    values = {"a": 5, "b": 7, "c": 3}

    expected_res = '<div style="display:none;" id="pouet">'\
        '<table class="popup"><tr><td>b</td><td class="value">'\
        '<img src="html/bar.png" width="100" height="16"/>7</td></tr><tr><td>a'\
        '</td><td class="value"><img src="html/bar.png" width="71" height="16"/>'\
        '5</td></tr><tr><td>c</td><td class="value"><img src="html/bar.png"'\
        ' width="42" height="16"/>3</td></tr></table></div>'

    bar_chart("pouet", values, doc)
    assert output.getvalue() == expected_res


def test_hamming_amb():
    """
    Test Hamming score to evaluation ambiguous loss
    """
    gold = [["A", "B", "C"],
           ["F", "G"]]

    predicted = [[{"A"}, {"E"}, {"C"}],
                [{"A"}, {"G"}]]

    assert hamming_loss(zip(gold, gold)) == 0.0
    assert hamming_loss(zip(gold, predicted)) == 2 / 5


def test_loss_1ex():
    
    gold = [["A", "B", "A", "C"]]
    pred = [["B", "A", "A", "C"]]
    poss = [[{"A", "B"}, {"A", "B"}, {"A"}, {"A", "C"}]]
    
    assert amb_hamming_loss(pred, gold, poss) == 1.0 / 3
    assert amb_hamming_loss(gold, gold, poss) == 1.0


def test_loss_2ex():
    
    gold = [["A", "B", "A", "C"],
            ["A", "A"]]
    pred = [["B", "A", "A", "C"],
            ["A", "B"]]
    poss = [[{"A", "B"}, {"A", "B"}, {"A"}, {"A", "C"}],
            [{"A", "B"}, {"A", "B"}]]

    assert amb_hamming_loss(pred, gold, poss) == 2.0 / 5
    assert amb_hamming_loss(gold, gold, poss) == 1.0
