import pytest

from sequence_labeling.laso_tagger import AveragedPerceptron, SkipPerceptron
from sequence_labeling.evaluation import hamming_loss


@pytest.fixture
def perceptron():
    weights = {"A": {"α": 10, "β": 1, "γ": 1},
               "B": {"α": 1, "β": 10, "γ": 1},
               "C": {"α": 1, "β": 1, "γ": 1}}

    p = AveragedPerceptron()
    p.weights = weights

    return p


@pytest.fixture
def skip_perceptron():
    weights = {"A": {"α": 10, "β": 1, "γ": 1},
               "B": {"α": 1, "β": 10, "γ": 1},
               "C": {"α": 1, "β": 1, "γ": 1}}

    p = SkipPerceptron()
    p.weights = weights

    return p


def test_pickle(perceptron):
    """
    Can we serialize a perceptron?
    """
    import pickle

    a = pickle.dumps(perceptron)
    new_p = pickle.loads(a)

    assert new_p.weights == perceptron.weights


def test_predict(perceptron):
    """
    Use the perceptron to predict a label
    """

    features = {"α": 1, "γ": 1}

    assert perceptron.predict(features) == "A"


def test_predict_restricted(perceptron):
    """
    Use the perceptron when the label set is restricted
    """
    features = {"α": 1, "γ": 1}

    # best label should be A ; but it is not in the set of authorized labels
    assert perceptron.predict(features, {"B", "C"}) == "C"

    # now A is in the set of authorized labels
    assert perceptron.predict(features, {"B", "A"}) == "A"

    # what happens when the set of authorized labels is empty
    assert perceptron.predict(features, {}) == ""


def test_no_update():

    p = AveragedPerceptron()

    p.update("A", "A", {})

    assert p.weights == {}


def test_simple_update(perceptron):

    perceptron.update("A", "C",  {"α": 1, "γ": 1})

    new_weights = {"A": {"α": 11, "β": 1, "γ": 2},
                   "B": {"α": 1, "β": 10, "γ": 1},
                   "C": {"α": 0, "β": 1, "γ": 0}}

    assert perceptron.weights == new_weights


def test_update_sequence(perceptron):

    perceptron.update("C", "C",  {"α": 1, "γ": 1})
    perceptron.update("C", "C",  {"α": 1, "γ": 1})
    perceptron.update("C", "A",  {"α": 1, "γ": 1})

    cum_weights = {"A": {"α": 29, "γ": 2},
                   "C": {"α": 4, "γ": 4}}

    new_weights = {"A": {"α": 9, "β": 1, "γ": 0},
                   "B": {"α": 1, "β": 10, "γ": 1},
                   "C": {"α": 2, "β": 1, "γ": 2}}

    assert perceptron.weights == new_weights
    assert perceptron._accum == cum_weights


def test_average_no_updates(perceptron):

    weights = perceptron.weights

    perceptron.update("C", "C",  {"α": 1, "γ": 1})
    perceptron.update("C", "C",  {"α": 1, "γ": 1})
    perceptron.update("C", "C",  {"α": 1, "γ": 1})

    perceptron.average_weights()

    assert perceptron.weights == weights


def test_average_updates(perceptron):
    """
    Average the weights when there are updates
    """

    perceptron.update("C", "C",  {"α": 1, "γ": 1})
    perceptron.update("C", "C",  {"α": 1, "γ": 1})
    perceptron.update("C", "A",  {"α": 1, "γ": 1})

    perceptron.average_weights()

    expected_weights = {"A": {"α": 29 / 3, "β": 1, "γ": 2 / 3},
                        "B": {"α": 1, "β": 10, "γ": 1},
                        "C": {"α": 4 / 3, "β": 1, "γ": 4 / 3}}

    assert perceptron.weights == expected_weights


def test_average_no_last_update(perceptron):
    """
    Average the weights when the last update has not yet been performed
    """

    perceptron.update("C", "C",  {"α": 1, "γ": 1})
    perceptron.update("C", "C",  {"α": 1, "γ": 1})
    perceptron.update("C", "A",  {"α": 1, "γ": 1})
    perceptron.update("C", "C",  {"α": 1, "γ": 1})

    perceptron.average_weights()

    expected_weights = {"A": {"α": 38 / 4, "β": 1, "γ": 2 / 4},
                        "B": {"α": 1, "β": 10, "γ": 1},
                        "C": {"α": 6 / 4, "β": 1, "γ": 6 / 4}}

    assert perceptron.weights == expected_weights


def test_real_learning():

    p = AveragedPerceptron()

    p.update("A", "B", {"α": 1, "β": 1, "γ": 1})
    p.update("B", "C", {"α": 1, "γ": 1})
    p.update("B", "B", {"α": 1, "β": 1})
    p.update("A", "B", {"α": 1, "β": 1, "γ": 1})
    p.update("A", "A", {"β": 1, "γ": 1})
    p.update("C", "B", {"γ": 1})

    expected_weights = {'A': {'α': 2.0, 'γ': 2.0, 'β': 2.0},
                        'C': {'α': -1.0, 'γ': 0.0},
                        'B': {'α': -1.0, 'γ': -2.0, 'β': -2.0}}

    assert p.weights == expected_weights


def test_hamming():

    gold = [["A", "B", "C"],
           ["F", "G"]]

    predicted = [["A", "E", "C"],
                ["A", "G"]]

    assert hamming_loss(zip(gold, gold)) == 0.0
    assert hamming_loss(zip(gold, predicted)) == 2 / 5


# =====================
# SkipPerceptron tests
# =====================


def test_skip_single_update_one_label(skip_perceptron):

    skip_perceptron.update({"A"}, "C", {"α": 1, "β": 1, "γ": 1})

    expected_weights = {"A": {"α": 11, "β": 2, "γ": 2},
                        "B": {"α": 1, "β": 10, "γ": 1},
                        "C": {"α": 0, "β": 0, "γ": 0}}

    assert skip_perceptron.weights == expected_weights


def test_skip_single_update_several_labels(skip_perceptron):
    """
    Update the SkipPerceptron when there is a single label
    """
    w = skip_perceptron.weights
    skip_perceptron.update({"A", "B"}, "C", {"α": 1, "β": 1, "γ": 1})

    assert skip_perceptron.weights == w


def test_skip_predict(skip_perceptron):
    """
    Prediction with a SkipPerceptron
    """

    features = {"α": 1, "γ": 1}

    assert skip_perceptron.predict(features) == "A"
