import pytest

from collections import defaultdict

from sequence_labeling.laso_tagger import POSTagger, AveragedPerceptron
from sequence_labeling.pos_tagger import pos_features


def test_tag_without_training():
    with pytest.raises(Exception):
        tagger = POSTagger(AveragedPerceptron)
        tagger.tag(["My", "name"])


def test_features():
    """
    Extract features in a simple case
    """

    context = defaultdict(lambda: None, enumerate(["a", "b", "c"]))

    expected_res = {
        "biais": 1,
        "tag@i-1=A": 1,
        "tag@i-2=None": 1,
        "tag@i-1=A_and_tag@i-2=None": 1,
        "tag@i-1=A_and_word@i=b": 1,
        "w@i+0=b": 1,
        "w@i-1=a": 1,
        "w@i-2=None": 1,
        "w@i+1=c": 1,
        "w@i+2=None": 1,
        'w@0=isLowerCase': 1,
        'suffix@i+0=b': 1,
        'suffix@i+1=c': 1,
        'suffix@i-1=a': 1}

    res = pos_features(1, context, ["A", None])
    assert res == expected_res


def test_features_real_example():
    import pickle
    import os

    dir_name = os.path.dirname(os.path.realpath(__file__))
    suffixes_file = os.path.join(dir_name, "suffixes_to_keep.pickle")

    words = ['επαvάληψη', 'ο', 'συvσδoυ']
    suffixes = pickle.load(open(suffixes_file, "rb"))
    context = defaultdict(lambda: None, enumerate(words))

    expected_features = {'biais': 1,
                         'suffix@i-1=ηψη': 1,
                         'suffix@i-1=ψη': 1,
                         'tag@i-1=NOUN_and_tag@i-2=None': 1,
                         'tag@i-2=None': 1,
                         'tag@i-1=NOUN_and_word@i=ο': 1,
                         'tag@i-1=NOUN': 1,
                         'w@0=isLowerCase': 1,
                         'w@i-1=επαvάληψη': 1,
                         'w@i-2=None': 1,
                         'w@i+0=ο': 1,
                         'w@i+1=συvσδoυ': 1,
                         'w@i+2=None': 1,
                         }

    res = pos_features(1, context, ["NOUN", None], suffixes)
    assert res == expected_features


def test_features_empty_suffixes():
    """
    Add all features when suffix dictionnary is empty
    """

    context = defaultdict(lambda: None, enumerate("bonjour les amis".split()))
    features = pos_features(2, context, ["A", None])
    expected_features = {'biais': 1,
                         'tag@i-1=A': 1,
                         'tag@i-1=A_and_tag@i-2=None': 1,
                         'tag@i-1=A_and_word@i=amis': 1,
                         'tag@i-2=None': 1,
                         'suffix@i+0=is': 1,
                         'suffix@i-1=es': 1,
                         'suffix@i-2=ur': 1,
                         'suffix@i+0=mis': 1,
                         'suffix@i-1=les': 1,
                         'suffix@i-2=our': 1,
                         'w@0=isLowerCase': 1,
                         'w@i+0=amis': 1,
                         'w@i+1=None': 1,
                         'w@i+2=None': 1,
                         'w@i-1=les': 1,
                         'w@i-2=bonjour': 1}

    assert features == expected_features


def test_filter_words():
    """
    Filter words in a simple case
    """

    context = defaultdict(lambda: None, enumerate("bonjour les amis".split()))
    features = pos_features(2, context, ["A", None], {}, {"bonjour", "amis"})
    expected_features = {'biais': 1,
                         'w@0=isLowerCase': 1,
                         'tag@i-1=A': 1,
                         'tag@i-1=A_and_tag@i-2=None': 1,
                         'tag@i-1=A_and_word@i=amis': 1,
                         'tag@i-2=None': 1,
                         'w@i+0=amis': 1,
                         'w@i+1=None': 1,
                         'w@i+2=None': 1,
                         'w@i-1=_NF_': 1,
                         'w@i-2=bonjour': 1}

    assert features == expected_features


def test_filter_current_word():
    """
    Filter word when current word is not frequent
    """

    context = defaultdict(lambda: None, enumerate("bonjour les amis".split()))
    features = pos_features(0, context, ["A", None], {}, {"les", "amis"})
    expected_features = {'biais': 1,
                         'w@0=isLowerCase': 1,
                         'tag@i-1=A': 1,
                         'tag@i-1=A_and_tag@i-2=None': 1,
                         'tag@i-1=A_and_word@i=_NF_': 1,
                         'tag@i-2=None': 1,
                         'w@i+0=_NF_': 1,
                         'w@i+1=les': 1,
                         'w@i+2=amis': 1,
                         'w@i-1=None': 1,
                         'w@i-2=None': 1}

    assert features == expected_features


def test_filter_digit():
    """
    Keep digits when filtering
    """
    context = defaultdict(lambda: None, enumerate("1000 les amis".split()))
    features = pos_features(0, context, ["A", None], {}, {"les", "amis"})
    expected_features = {'biais': 1,
                         'digitInWord': 1,
                         'w@0=isLowerCase': 1,
                         'tag@i-1=A': 1,
                         'tag@i-1=A_and_tag@i-2=None': 1,
                         'tag@i-1=A_and_word@i=!DIGIT': 1,
                         'tag@i-2=None': 1,
                         'w@i+0=!DIGIT': 1,
                         'w@i+1=les': 1,
                         'w@i+2=amis': 1,
                         'w@i-1=None': 1,
                         'w@i-2=None': 1}

    assert features == expected_features


def test_uppercase_letter():
    """
    Can we detect when the word starts with an upper letter
    """

    context = defaultdict(lambda: None, enumerate("bonjour les Amis".split()))
    features = pos_features(2, context, ["A", None], {}, {"bonjour", "amis"})
    expected_features = {'biais': 1,
                         'w@0=isUpperCaseAndNotFirst': 1,
                         'tag@i-1=A': 1,
                         'tag@i-1=A_and_tag@i-2=None': 1,
                         'tag@i-1=A_and_word@i=amis': 1,
                         'tag@i-2=None': 1,
                         'w@i+0=amis': 1,
                         'w@i+1=None': 1,
                         'w@i+2=None': 1,
                         'w@i-1=_NF_': 1,
                         'w@i-2=bonjour': 1}

    assert features == expected_features


def test_upper_at_beginning():
    context = defaultdict(lambda: None, enumerate("Bonjour les amis".split()))
    features = pos_features(0, context, [None, None], {}, {"bonjour", "amis"})
    expected_features = {'biais': 1,
                         'w@0=isLowerCase': 1,
                         'tag@i-1=None': 1,
                         'tag@i-1=None_and_tag@i-2=None': 1,
                         'tag@i-1=None_and_word@i=bonjour': 1,
                         'tag@i-2=None': 1,
                         'w@i+0=bonjour': 1,
                         'w@i+1=_NF_': 1,
                         'w@i+2=amis': 1,
                         'w@i-1=None': 1,
                         'w@i-2=None': 1}

    assert features == expected_features


def test_digit():
    """
    Are we able to recognize digits
    """
    context = defaultdict(lambda: None, enumerate("à 12h32 précise".split()))
    features = pos_features(1, context, ["B", "C"], {}, None)
    expected_features = {'tag@i-1=B_and_word@i=12h32': 1,
                         'w@i+2=None': 1,
                         'w@i+1=précise': 1,
                         'w@i+0=12h32': 1,
                         'w@i-1=à': 1,
                         'w@0=isLowerCase': 1,
                         'biais': 1,
                         'tag@i-1=B_and_tag@i-2=C': 1,
                         'digitInWord': 1,
                         'tag@i-1=B': 1,
                         'w@i-2=None': 1,
                         'tag@i-2=C': 1}

    assert features == expected_features

    context = defaultdict(lambda: None, enumerate("les 100 meilleurs".split()))
    features = pos_features(1, context, ["B", "C"], {}, None)
    expected_features = {'tag@i-1=B_and_word@i=!DIGIT': 1,
                         'w@i+2=None': 1,
                         'w@i+1=meilleurs': 1,
                         'w@i+0=!DIGIT': 1,
                         'w@i-1=les': 1,
                         'w@0=isLowerCase': 1,
                         'biais': 1,
                         'tag@i-1=B_and_tag@i-2=C': 1,
                         'digitInWord': 1,
                         'tag@i-1=B': 1,
                         'w@i-2=None': 1,
                         'tag@i-2=C': 1}

    assert features == expected_features


def test_features_with_types():
    """
    Extract features when type information is given
    """

    type_constraints = lambda x: [x.upper(), "X"]
    context = defaultdict(lambda: None, enumerate("a b c d".split()))
    features = pos_features(1, context, ["B", "C"], type_dict=type_constraints)
    
    expected_features = {'w@i-1=a': 1,
                         'suffix@i+2=d': 1,
                         'tag@i-1=B': 1,
                         'w@i+2=d': 1,
                         'w@i-2=None': 1,
                         'suffix@i-1=a': 1,
                         'w@0=isLowerCase': 1,
                         'tag@i-1=B_and_tag@i-2=C': 1,
                         'biais': 1,
                         'tag@i-1=B_and_word@i=b': 1,
                         'suffix@i+1=c': 1,
                         'w@i+0=b': 1,
                         'suffix@i+0=b': 1,
                         'tag@i-2=C': 1,
                         'w@i+1=c': 1,
                         'word@+0_has_label_B': 1,
                         'word@+0_has_label_X': 1,
                         'word@+1_has_label_C': 1,
                         'word@+1_has_label_X': 1,
                         'word@+2_has_label_D': 1,
                         'word@+2_has_label_X': 1,
                         'word@-1_has_label_A': 1,
                         'word@-1_has_label_X': 1}

    assert features == expected_features


def test_features_with_types():
    """
    Extract features when type information is given but not for all words.
    """

    type_constraints = lambda x: [x.upper(), "X"] if x != "a" else []
    context = defaultdict(lambda: None, enumerate("a b c d".split()))
    features = pos_features(1, context, ["B", "C"], type_dict=type_constraints)
    
    expected_features = {'w@i-1=a': 1,
                         'suffix@i+2=d': 1,
                         'tag@i-1=B': 1,
                         'w@i+2=d': 1,
                         'w@i-2=None': 1,
                         'suffix@i-1=a': 1,
                         'w@0=isLowerCase': 1,
                         'tag@i-1=B_and_tag@i-2=C': 1,
                         'biais': 1,
                         'tag@i-1=B_and_word@i=b': 1,
                         'suffix@i+1=c': 1,
                         'w@i+0=b': 1,
                         'suffix@i+0=b': 1,
                         'tag@i-2=C': 1,
                         'w@i+1=c': 1,
                         'word@+0_has_label_B': 1,
                         'word@+0_has_label_X': 1,
                         'word@+1_has_label_C': 1,
                         'word@+1_has_label_X': 1,
                         'word@+2_has_label_D': 1,
                         'word@+2_has_label_X': 1}

    assert features == expected_features


def test_features_history():
    """
    Do we correctly consider the size of the history?
    """

    context = defaultdict(lambda: None, enumerate("a b c d".split()))
    expected_features = {'biais': 1,
                         'suffix@i+0=d': 1,
                         'suffix@i-1=c': 1,
                         'suffix@i-2=b': 1,
                         'tag@i-1=B': 1,
                         'tag@i-1=B_and_tag@i-2=C': 1,
                         'tag@i-1=B_and_word@i=d': 1,
                         'tag@i-2=C': 1,
                         'w@0=isLowerCase': 1,
                         'w@i+0=d': 1,
                         'w@i+1=None': 1,
                         'w@i+2=None': 1,
                         'w@i-1=c': 1,
                         'w@i-2=b': 1}

    assert pos_features(3, context, ["B", "C"]) == expected_features

    # larger history but the history size is kept unchanged
    assert pos_features(3, context, ["B", "C", "A"]) == expected_features

    expected_features = {'biais': 1,
                         'suffix@i+0=d': 1,
                         'suffix@i-1=c': 1,
                         'suffix@i-2=b': 1,
                         'tag@i-1=B': 1,
                         'tag@i-1=B_and_tag@i-2=C': 1,
                         'tag@i-1=B_and_word@i=d': 1,
                         'tag@i-2=C': 1,
                         'tag@i-3=A': 1,
                         'w@0=isLowerCase': 1,
                         'w@i+0=d': 1,
                         'w@i+1=None': 1,
                         'w@i+2=None': 1,
                         'w@i-1=c': 1,
                         'w@i-2=b': 1}
    # larger history & larger history size to consider
    features = pos_features(3, context, ["B", "C", "A"], history_size=3)
    assert features == expected_features

    # history smaller than the history size
    features = pos_features(0, context, [], history_size=3)
    expected_features = {'biais': 1,
                         'suffix@i+0=a': 1,
                         'suffix@i+1=b': 1,
                         'suffix@i+2=c': 1,
                         'tag@i-1=None': 1,
                         'tag@i-1=None_and_tag@i-2=None': 1,
                         'tag@i-1=None_and_word@i=a': 1,
                         'tag@i-2=None': 1,
                         'tag@i-3=None': 1,
                         'w@0=isLowerCase': 1,
                         'w@i+0=a': 1,
                         'w@i+1=b': 1,
                         'w@i+2=c': 1,
                         'w@i-1=None': 1,
                         'w@i-2=None': 1}
    assert features == expected_features
