# Copyright (c) 2014, G. Wisniewski
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""
Build the ressources necessary to apply [Li et al., 2012] features filter.

- only the token that appears more than 10 times are considered in the
   features.
- only the suffixes that appears in more that 20 *types* are considered in
  the features

This script extracts the set of suffixes and of tokens to consider.
"""
import pickle

from collections import Counter

from weakly_io import pickle_reader
from sequence_labeling.pos_tagger import normalize, YEAR, DIGIT


def create_feature_template(corpus_fn, frequent_words_fn,
                            frequent_suffixes_fn, listener=None):
    """
    Parameters
    ----------
    - corpus_fn, a string
      the name of the pickle file in which the corpus is stored.
    - frequent_words_fn, a string
      the name of the file in which the frequent words are going to be
      stored
    - frequent_suffixes_fn, a string
      the name of the file in which the frequent suffixes are going to
      be stored
    """
    input = open(corpus_fn, "rb")
    words_count = Counter()

    for line in listener.iter(pickle_reader(input)):
        words_count.update(normalize(w.lower(), None) for w in line["tgt"])

    suffixes_count = Counter()
    for word in words_count:

        if word in {YEAR, DIGIT}:
            continue

        if len(word) >= 2:
            suffixes_count.update([word[-2:]])

        if len(word) >= 3:
            suffixes_count.update([word[-3:]])

    suffixes = {s for s, count in suffixes_count.items() if count >= 20}
    words = {w for w, count in words_count.items() if count >= 10}

    with open(frequent_words_fn, "wb") as ofile:
        listener.msg("keeping {:,} words out of {:,}".format(
                len(words), len(words_count)))
        pickle.dump(words, ofile, pickle.HIGHEST_PROTOCOL)

    with open(frequent_suffixes_fn, "wb") as ofile:
        listener.msg("keeping {:,} suffixes out of {:,}".format(
                len(suffixes), len(suffixes_count)))
        pickle.dump(suffixes, ofile, pickle.HIGHEST_PROTOCOL)
