import pickle
import logging

from operator import itemgetter
from collections import defaultdict

import frogress

from weakly_io import prepare_test_data, read_test

from sequence_labeling.evaluation import make_report
from sequence_labeling.laso_tagger import POSTagger, WeakilySupervisedPerceptron
from sequence_labeling.tools import listify
from sequence_labeling.pos_tagger import pos_features as pfeatures

@listify
def label(data, tagger, quiet=True):

    if not quiet:
        logging.info("predicting labels")
        bar = frogress.bar
    else:
        bar = lambda x: (xx for xx in x)

    for words, gold, ambig in bar(data):
        predicted_labels = tagger.tag(words, ambig)

        yield words, gold, predicted_labels, ambig


def decorate_data(words, gold_labels, pred_labels, ambig_labels, pos_features):

    errors = [gold != pred for gold, pred in zip(gold_labels, pred_labels)]
    amb_errors = [g not in amb for g, amb in zip(gold_labels, ambig_labels)]

    ctx = defaultdict(lambda: None, enumerate(words))

    features = []
    prev, prev_prev = None, None
    for i, label in enumerate(pred_labels):
        features.append(pos_features(i, ctx, [prev, prev_prev]))
        prev_prev = prev
        prev = label

    scores = [tagger.model.labels_scores(f) for f in features]

    res_features = []
    w = tagger.model.weights
    for feat in features:
#        scored_features = ((f + "&" + c, w[c][f]) for f in feat for c in w)
#        scored_features = sorted(scored_features, key=itemgetter(1),
#                                 reverse=True)
#        scored_features = dict(scored_features[:30])
        scored_features = {"empty": 1}
        res_features.append(scored_features)

    return {"gold": gold_labels,
            "errors": errors,
            "amb_errors": amb_errors,
            "features": res_features,
            "words": words,
            "ambiguous_label": ambig_labels,
            "labels_scores": scores,
            "pred": pred_labels}


if __name__ == "__main__":

    import argparse

    from functools import partial

    parser = argparse.ArgumentParser()
    parser.add_argument("--model", required=True,
                        type=argparse.FileType("rb"))
    parser.add_argument("--corpus", required=True)
    parser.add_argument("--output", default="output.html",
                        type=argparse.FileType("wt"))

    args = parser.parse_args()


    logging.info("logging tagger...")
    tagger = pickle.load(args.model)

    data = prepare_test_data(read_test(args.corpus),
                             wiki_constraints={},
                             alignment_constraints={},
                             priority_constraints={})
    data = label(data, tagger)

    logging.info("decorating data...")
    ddata = partial(decorate_data, pos_features=tagger.features)
    decorated_data = [ddata(*d) for d in frogress.bar(data)]

    logging.info("generating html")
    make_report(decorated_data, ["metrics", "f1_table", "sentences"],
                args.output)
    #make_report(decorated_data, ["dump_errors"], args.output)


