import warnings
import pickle
import logging
import json

from functools import partial

from sequence_labeling.pos_tagger import pos_features
from project_labels import type_constraints_intersection, type_constraints_union

from weakly_tasks import *


logging.basicConfig(level=logging.INFO)

n_ex = 100000
history_size = 3
weakly_n_iter = 1


logging.basicConfig(level=logging.DEBUG)

def get_type_constraints(lang, which):
    wiki_constraints = WIKI_CONSTRAINTS.format(lang) 
    priority_constraints = PRIORITY_CONSTRAINTS.format(lang, priority_constraints)

    priority_constraints = read_priority_constraints(priority_constraints)
    wiki_constraints = read_type_constraint(open(wiki_constraints, "rt"))
    al_constraints = json.load(open(AL_CONSTRAINTS.format(lang), "rt"))

    if which == "union":
        type_constraints = partial(type_constraints_union,
                                   wiki_constraints=wiki_constraints,
                                   al_constraints=al_constraints,
                                   priority_constraints=priority_constraints)
    elif which == "intersection":
        type_constraints = partial(type_constraints_intersection,
                                   wiki_constraints=wiki_constraints,
                                   al_constraints=al_constraints,
                                   priority_constraints=priority_constraints)
    else:
        raise Exception

    return type_constraints


def make_type_constraints(lang, priority_constraints):
    warnings.warn("make_type_constraints should not be used", DeprecationWarning)
    wiki_constraints = WIKI_CONSTRAINTS.format(lang) 
    priority_constraints = PRIORITY_CONSTRAINTS.format(lang, priority_constraints)

    priority_constraints = read_priority_constraints(priority_constraints)
    wiki_constraints = read_type_constraint(open(wiki_constraints, "rt"))
    al_constraints = json.load(open(AL_CONSTRAINTS.format(lang), "rt"))

    type_constraints = partial(type_constraints_intersection,
                               wiki_constraints=wiki_constraints,
                               al_constraints=al_constraints,
                               priority_constraints=priority_constraints)
    return type_constraints


def make_feature_template(lang, h_size, constraints):
    """
    Create the feature template for POS tagging

    Paramaters
    ----------

    - lang, a string
      the language considered (to load the appropriate resources)
    - h_size, an int
      the number of past decisions to consider as features
    - constraints, a callable
      returns the set of allowed POS tags for a word. If `None` these
      features are not considered
    """

    if constraints:
        type_constraints = constraints
    else:
        type_constraints = None
        
    words = pickle.load(open(FREQUENT_WORDS.format(lang), "rb"))
    suffixes = pickle.load(open(FREQUENT_SUFFIXES.format(lang), "rb"))

    features = partial(pos_features,
                       allowed_suffixes=suffixes,
                       allowed_words=words,
                       history_size=h_size,
                       type_dict=type_constraints)

    return features


# =======================
# SUPERVISED EXPERIMENTS
# =======================

def exp_super_1(lang, job_queue):
    """
    Experiment n°1
    --------------
    
    train a baseline POS tagger (supervised)
    """
    logging.info(exp_super_1.__doc__)
    logging.info("lang: {}".format(lang))
        
    train_set = CONLL_TRAIN.format(lang)
    output = SUPER_MODEL.format(lang=lang)
    train_job = job_queue.enqueue_call(train_supervised_tagger,
                                       (lang, train_set, "basic_features", output),
                                       track_progress=True)

    test_set = CONLL_TEST.format(lang)

    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang),
                                 depends_on=train_job)
    
    if job.result is not None:
        logging.info("[res] {},supervised_baseline,{}".format(lang, job.result))

    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang, "full"),
                                 depends_on=train_job)
    if job.result is not None:
        logging.info("[res] {},supervised_baseline_type_during_inference,{}".format(lang, job.result))


def exp_super_2(lang, job_queue):
    """
    Experiment n°2
    --------------

    Consider features describing type constraints for supervised POS tagger
    """
    logging.info(exp_super_2.__doc__)
    logging.info("lang: {}".format(lang))

    train_set = CONLL_TRAIN.format(lang)
    tagger = SUPER_MODEL_WITH_TYPES.format(lang=lang)
    train_job = job_queue.enqueue_call(train_supervised_tagger,
                                       (lang, train_set, "constraints_features", tagger), 
                                       track_progress=True)

    test_set = CONLL_TEST.format(lang)
    job = job_queue.enqueue_call(test_tagger,
                                 (tagger, test_set, lang),
                                 depends_on=train_job)
    if job.result is not None:
        logging.info("[res] {},supervised_type_as_features,{}".format(lang, job.result))

    job = job_queue.enqueue_call(test_tagger,
                                 (tagger, test_set, lang, "full"),
                                 depends_on=train_job)
    if job.result is not None:
        logging.info("[res] {},supervised_type_as_features_and_inference,{}".format(lang, job.result))


def exp_super_3(lang, job_queue):
    """
    Supervised Experiment n°3
    --------------------------
    
    train a supervised POS tagger when labels are constrained during training
    """
    logging.info(exp_super_3.__doc__)
    logging.info("lang: {}".format(lang))
        
    train_set = CONLL_TRAIN.format(lang)
    output = SUPER_MODEL_CONSTRAINED.format(lang=lang)
    train_job = job_queue.enqueue_call(train_supervised_tagger,
                                       (lang, train_set, "basic_features", output, "full"),
                                       track_progress=True)
    print("*** {}".format(train_job.result))
    test_set = CONLL_TEST.format(lang)

    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang),
                                 depends_on=train_job)
    
    if job.result is not None:
        logging.info("[res] {},supervised_baseline_constrained,{}".format(lang, job.result))

    job = job_queue.enqueue_call(test_tagger,
                               (output, test_set, lang, "full"),
                               depends_on=train_job)
    if job.result is not None:
        logging.info("[res] {},supervised_baseline_constrained_type_during_inference,{}".format(lang, job.result))


# =============================
# WEAKLY SUPERVISED EXPERIMENTS
# =============================

def exp3(lang, job_queue):
    """
    Experiment n°3
    --------------

    Train a weakly supervised POS tagger with the standard feature set.
    """
    logging.info(exp3.__doc__)
    logging.info("lang: {}".format(lang))

    train_set = WEAKLY_INTER_TRAIN.format(lang=lang)
    output = WEAKLY_INTER_MODEL.format(lang)
    train_job = job_queue.enqueue_call(train_weakly_tagger,
                                       (lang, train_set, "basic_features", output),
                                       track_progress=True,
                                       name="[{}] train baseline weakly inter".format(lang))

    test_set = CONLL_TEST.format(lang)
    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang),
                                 depends_on=train_job,
                                 name="[{}] test baseline weakly inter".format(lang))
    if job.result is not None:
        logging.info("[res] {},baseline_weakly_inter,{}".format(lang, job.result))

    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang, "full"),
                                 depends_on=train_job,
                                 name="[{}] test baseline weakly inter type during test".format(lang))

    if job.result is not None:
        logging.info("[res] {},baseline_weakly_inter_type_during_test,{}".format(lang, job.result))

    # =================================
    # Use type constraints as features
    # =================================
    output = WEAKLY_INTER_MODEL_EXTEND_FEATURES.format(lang)
    train_job_extended = job_queue.enqueue_call(train_weakly_tagger,
                                                (lang, train_set, "constraints_features", output),
                                                track_progress=True,
                                                name="[{}] train extended weakly inter".format(lang))
                                 

    test_set = CONLL_TEST.format(lang)
    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang),
                                 depends_on=train_job_extended,
                                 name="[{}] test extended weakly inter".format(lang))
    if job.result is not None:
        logging.info("[res] {},extended_weakly_inter,{}".format(lang, job.result))

    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang, "full"),
                                 depends_on=train_job_extended,
                                 name="[{}] test extended weakly inter type during test".format(lang))
    if job.result is not None:
        logging.info("[res] {},extended_weakly_inter_type_during_test,{}".format(lang, job.result))


def exp4(lang, job_queue):
    """
    Experiment n°4
    --------------

    Train a weakly supervised POS tagger with the standard feature set
    and constraint label during training (aka the `Täckström' way)
    """
    logging.info(exp4.__doc__)
    logging.info("lang: {}".format(lang))

    train_set = WEAKLY_INTER_TRAIN.format(lang=lang)
    output = WEAKLY_INTER_MODEL_WITH_TYPE.format(lang)

    train_job = job_queue.enqueue_call(train_weakly_tagger,
				       (lang, train_set,
				       "basic_features", output, "full"),
				       track_progress=True)

    test_set = CONLL_TEST.format(lang)
    r = job_queue.enqueue_call(test_tagger,
                               (output, test_set, lang),
                               depends_on=train_job)
    if r.result is not None:
        logging.info("[res] {},weakly_train_with_type,{}".format(lang, r.result))

    r = job_queue.enqueue_call(test_tagger,
                               (output, test_set, lang, "full"),
                               depends_on=train_job)
    if r.result is not None:
        logging.info("[res] {},weakly_baseline_type_during_train_and_inference,{}".format(lang, r.result))


def exp5(lang, job_queue):
    """
    Experiment n°5
    --------------
    
    Train a weakly supervised POS tagger using type constraints as features
    """
    logging.info(exp5.__doc__)
    logging.info("lang: {}".format(lang))

    train_set = WEAKLY_INTER_TRAIN.format(lang=lang)
    output = WEAKLY_INTER_MODEL_WITH_TYPE_FEATURES.format(lang)
    
    train_job = job_queue.enqueue_call(train_weakly_tagger,
                                       (lang, train_set,
                                        "constraints_features", output),
                                       track_progress=True)
    
    test_set = CONLL_TEST.format(lang)
    r = job_queue.enqueue_call(test_tagger,
                               (output, test_set, lang),
                               depends_on=train_job)
    if r.result is not None:
        logging.info("{},weakly_with_type_features,{}".format(lang, r.result))

    r = job_queue.enqueue_call(test_tagger,
                               (output, test_set, lang, "full"),
                               depends_on=train_job)
    
    if r.result is not None:
        logging.info("{},weakly_with_type_features_and_inference,{}".format(lang, r.result))


def exp6(lang, job_queue):
    """
    Experiment n°6
    --------------

    Train a weakly supervised POS tagger skipping ambiguous examples
    """

    logging.info(exp6.__doc__)
    logging.info("lang: {}".format(lang))

    train_set = WEAKLY_INTER_TRAIN.format(lang=lang)
    output = SKIP_WEAKLY_MODEL.format(lang)
    
    train_job = job_queue.enqueue_call(train_weakly_tagger,
                                       (lang, train_set,
                                       "basic_features", output, None, True),
                                       track_progress=True,
                                       name="[{}] train skip".format(lang))

    test_set = CONLL_TEST.format(lang)
    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang),
                                 depends_on=train_job,
                                 name="[{}] test inter skip".format(lang))

    if job.result is not None:
        logging.info("{},baseline_inter_skip,{}".format(lang, job.result))

    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang, "full"),
                                 depends_on=train_job,
                                 name="[{}] test inter skip + constraints".format(lang))

    if job.result is not None:
        logging.info("{},baseline_inter_skip_constraints,{}".format(lang, job.result))

    # =============
    # SKIP + UNION
    # =============
    train_set = WEAKLY_UNION_TRAIN.format(lang=lang)
    output = SKIP_WEAKLY_UNION_MODEL.format(lang)
    
    train_job = job_queue.enqueue_call(train_weakly_tagger,
                                       (lang, train_set,
                                       "basic_features", output, None, True),
                                       track_progress=True,
                                       name="[{}] train union skip".format(lang))


    test_set = CONLL_TEST.format(lang)
    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang),
                                 depends_on=train_job,
                                 name="[{}] test skip".format(lang))

    if job.result is not None:
        logging.info("{},baseline_union_skip,{}".format(lang, job.result))


def exp7(lang):
    """
    Experiment 7
    ------------

    Direct evaluation of POS tagging performance: performance of a
    dependency parser

    We run the following test:
    - performance of a dependency parser considering GOLD POS tags
    - performance of a dependency parser considering PREDICTED POS
      tags (with a **supervised** POS tagger)
    - performance of a dependency parser considering PREDICTED POS
      tags (with a **weakly** POS tagger)
    """
    import sys
    sys.path.append("../dependency_parser")

    import frogress
    from dependency_parser import full_dp_features,\
        train_dependency_parser, Parser, check_projectivity, \
        read_conll, test_parser

    def listener(what):

        if isinstance(what, str):
            print(what)
        else:
            return frogress.bar(what)

    for l in lang:
        # Supervised training
        # ===================

        filename = UDT_TRAIN.format(l)
        fh = open(filename, "rt", encoding="utf-8")
        fh = read_conll(fh.read())
        train_set = filter(lambda s: check_projectivity(s[2]), fh)
        train_set = list(train_set)

#         logging.info("train a DP parser from GOLD POS")
#         parser = Parser(full_dp_features)
#         train_dependency_parser(parser, train_set, 10, listener)

#         model = DP_MODEL.format(l)
#         pickle.dump(parser, open(model, "wb"), pickle.HIGHEST_PROTOCOL)

#         # Supervised testing
#         # ==================

        filename = UDT_TEST.format(l)
        fh = open(filename, "rt", encoding="utf-8")
        test_set = list(read_conll(fh.read()))

#         logging.info("evaluation DP parser with predicted POS")
#         r = test_parser(parser, test_set, True)
#         logging.info("{},dp_parser_pred_during_test,{}".format(l, r))

#         r = test_parser(parser, test_set, False)
#         logging.info("{},dp_parser_gold_during_test,{}".format(l, r))

        # Weakly training
        # ---------------
        tagger_fn = WEAKLY_MODEL.format(l)
        tagger = pickle.load(open(tagger_fn, "rb"))

        train_set = [(t[0], ["<start>"] + tagger.tag(t[0][1:-1]) + ["ROOT"], t[2], t[3]) for t in train_set]

        logging.info("train a DP parser from WEAKLY POS")
#        parser = Parser(full_dp_features)
#        train_dependency_parser(parser, train_set, 10, listener)

        model = DP_MODEL_WEAK_POS.format(l)
#        pickle.dump(parser, open(model, "wb"), pickle.HIGHEST_PROTOCOL)

        parser = pickle.load(open(model, "rb"))

        logging.info("evaluation DP parser with predicted POS")
        test_set = [(t[0], ["<start>"] + tagger.tag(t[0][1:-1]) + ["ROOT"], t[2], t[3]) for t in test_set]
        r = test_parser(parser, test_set, False)
        logging.info("{},dp_parser_pred_during_test,{}".format(l, r))


def exp8(lang, job_queue):
    """
    Experiment n°8
    --------------

    Train a weakly supervised POS tagger with the standard feature set
    using the union heuristic.
    """
    logging.info(exp8.__doc__)
    logging.info("lang: {}".format(lang))

    output = WEAKLY_UNION_MODEL.format(lang)
    train_set = WEAKLY_UNION_TRAIN.format(lang=lang)

    train_job = job_queue.enqueue_call(train_weakly_tagger,
                                       (lang, train_set, "basic_features", output),
                                       track_progress=True)

    test_set = CONLL_TEST.format(lang)
    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang),
                                 depends_on=train_job)
    if job.result is not None:
        logging.info("[res] {},weakly_union,{}".format(lang, job.result))

    job = job_queue.enqueue_call(test_tagger,
                                 (output, test_set, lang, "full"),
                                 depends_on=train_job)
    if job.result is not None:
        logging.info("[res] {},weakly_union_type_during_inference,{}".format(lang, job.result))


def exp_learning_curve(lang, job_queue):
    """
    Draw learning curve
    -------------------

    """
    logging.info(exp_learning_curve.__doc__)
    logging.info("lang: {}".format(lang))

    train_set = WEAKLY_INTER_TRAIN.format(lang=lang)
    for n_ex in [100, 1000, 10000, 100000, 1000000, 10000000]:
        output = WEAKLY_INTER_MODEL_LEARNING_CURVE.format(lang, n_ex)
        train_job = job_queue.enqueue_call(train_weakly_tagger,
                                           (lang, train_set, "basic_features", output),
                                           {"n_ex": n_ex},
                                           track_progress=True,
                                           name="[{}] train baseline weakly inter with n_ex = {:,}".format(lang, n_ex))
        
        test_set = CONLL_TEST.format(lang)

        # Evaluation with Hamming loss
        # ============================
        job = job_queue.enqueue_call(test_tagger,
                                     (output, test_set, lang),
                                     depends_on=train_job,
                                     name="[{}] test baseline weakly inter with n_ex = {:,}".format(lang, n_ex))
        if job.result is not None:
            logging.info("[res] {},baseline_weakly_inter,hamming,{},{}".format(lang, n_ex, job.result))
            
        job = job_queue.enqueue_call(test_tagger,
                                     (output, test_set, lang, "full"),
                                     depends_on=train_job,
                                     name="[{}] test baseline weakly inter type during test".format(lang))        
        if job.result is not None:
            logging.info("[res] {},baseline_weakly_inter_type_during_test,hamming,{},{}".format(lang, n_ex,job.result))

        # Evaluation with Ambiguous Hamming loss
        # ======================================
        job = job_queue.enqueue_call(test_tagger,
                                     (output, test_set, lang),
                                     {"metric": "amb_hamming"},
                                     depends_on=train_job,
                                     name="[{}] test baseline weakly inter with n_ex = {:,}".format(lang, n_ex))
        if job.result is not None:
            logging.info("[res] {},baseline_weakly_inter,amb_hamming,{},{}".format(lang, n_ex, job.result))
            
        job = job_queue.enqueue_call(test_tagger,
                                     (output, test_set, lang, "full"),
                                     {"metric": "amb_hamming"},
                                     depends_on=train_job,
                                     name="[{}] test baseline weakly inter type during test".format(lang))
        
        if job.result is not None:
            logging.info("[res] {},baseline_weakly_inter_type_during_test,amb_hamming,{},{}".format(lang, n_ex,job.result))


if __name__ == "__main__":
    from job_queue.job_queue import Queue, create_connection, FakeQueue
    job_queue = Queue(create_connection())
    
    # =================
    # Prepare datasets
    # =================
    
    lang = ["cs", "de", "el", "es", "fi", "fr", "id", "it", "sv"]



    for l in lang:
        # exp_super_1(l, job_queue)
        # exp_super_2(l, job_queue)
        #        exp_super_3(l, job_queue)
        #        exp3(l, job_queue)
        #         exp4(l, job_queue)
        #         exp5(l, job_queue)
        exp6(l, job_queue)
        # exp7(lang, q)
        #         exp8(l, job_queue)
        
    # exp_learning_curve("fr", job_queue)

