# Copyright (c) 2014, G. Wisniewski
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
"""
Infer the POS tags information of corpus by considering two kinds of
information:

- the “type” constraints (i.e. constraints specific to a type)
- the “token” constraints (i.e. constraints specific to a given token)

Constraints can be think of a mapping between a word and a set of possible POS
tags.

Token constraints are extracted from alignments.

Type constraints are extracted from:

- wiktionary by parsing the page of a word and extracting all POS-related
  information.
- alignment by considering all the POS tags a word is every aligned with.
"""

import logging
import pickle
import json

from functools import partial

from sequence_labeling.tools import listify
from weakly_io import pickle_reader


logging.basicConfig(level=logging.INFO)

# !!!Fix: structure is a bit weird now. I think we could restructure
# the code having a sing type constraints to apply, which is first
# computed (e.g. using union and intersection). This would make the
# code more reusable (any type constraints could be provided
# (e.g. only one, three or more, etc.). How does GW compute with only
# al_constraints ?


def get_type_constraints(word,
                         type_constraints,
                         priority_constraints=dict(),
                         lower_backoff=False):
    """
    Return allowed tags for a given word type `word`.

    If `priority_constraints` is given those constraints apply as a priority.

    If `lower_backoff` is true, `type_constraints` is also searched for a
    lowercase version of `word` if cased fails.

    """

    for word_template in priority_constraints:
        if word_template.fullmatch(word):
            return priority_constraints[word_template]

    if word in type_constraints:
        return type_constraints[word]

    if lower_backoff and word.lower() in type_constraints:
        return type_constraints[word.lower()]

    return set()


@listify
def apply_type_constraints(sentence,
                           wiki_constraints,
                           al_constraints,
                           priority_constraints,
                           merge="intersection"):
    """
    Label all words in a sentence by applying the different type constraints.

    There are three kinds of type constraints:
    - constraints extracted from wiktionary (`wiki_constraints`)
    - constraints extraced from all alignments (`al_constraints`(
    - hand-made constraints (`priority_constraints`) that can be used, for
      instance, to account for differences in conventions and punctuation

    Parameters
    ----------
    - sentence, a list of words
      the tokenized sentence
    - wiki_constraints, a mapping between a string and a list
      the constraints extracted from wiktionary
    - alignment_constraints, a mapping between a string and a list
      the constraints extracted from the alignment
    - priority_constraints, a mapping
      the priority constraints
    """

    if merge == "intersection":
        type_constraints_heuristic = type_constraints_intersection
    elif merge == "union":
        type_constraints_heuristic = type_constraints_union
    else:
        raise Exception()

    f = partial(type_constraints_heuristic,
                wiki_constraints=wiki_constraints,
                al_constraints=al_constraints,
                priority_constraints=priority_constraints)

    return map(f, sentence)


def type_constraints_intersection(word,
                                  wiki_constraints,
                                  al_constraints,
                                  priority_constraints,
                                  lower_backoff=True):
    """Return type constraints for `word` using the intersection of
    `wiki_constraints` and `al_constraints`.

    For each word type, if the intersection is empty, return first the
    constraints from `wiki_constraints` if not also empty and only then the one
    from `al_constraints`.

    Note that the lower cased backoff is applied independently for both
    constraints before attempting to perform the intersection.

    """

    al = get_type_constraints(word, al_constraints,
                              priority_constraints, lower_backoff)
    wiki = get_type_constraints(word, wiki_constraints,
                                priority_constraints, lower_backoff)

    if wiki.intersection(al):
        return wiki.intersection(al)
    elif wiki:
        return wiki
    else:
        return al


def type_constraints_union(word,
                           wiki_constraints,
                           al_constraints,
                           priority_constraints,
                           lower_backoff=True):
    """Return type constraints for `word` using the union of
    `wiki_constraints` and `al_constraints`.

    Note that the lower cased backoff is applied independently for both
    constraints before performing the union.

    """

    al = get_type_constraints(word, al_constraints,
                              priority_constraints, lower_backoff)
    wiki = get_type_constraints(word, wiki_constraints,
                                priority_constraints, lower_backoff)

    return wiki.union(al)


@listify
def apply_token_constraints(labeled_sentence, example):
    """Filter a sequence of ambiguous labels by applying the token
    constraints using the fusion method from [Täckström et al, 2013].

    This method assumes that the type constraints have already been
    applied.

    Parameters
    ----------

    - labels, a list of sets
         labels[i] is the i-th ambiguous supervision set for the i-th
         word. If no type constraints are available the set *must* be
         empty.

    - example, a mapping describing the associated example
         This mapping must define (at least) the keys:

         - al (the alignment)
         - src_pos (a list of strings describing the source POS tags)
    """

    # !!!Fix: wouldn't it be better to raise an Exception here if the
    # alignment is not one-to-one ?
    al = {tgt_index[0]: src_index for src_index, tgt_index
          in example["al"].items() if len(tgt_index) == 1}

    for index, label_set in enumerate(labeled_sentence):

        if index not in al:
            yield label_set
            continue

        token_label = example["src_pos"][al[index]]

        if token_label in label_set:
            yield {token_label}
        elif not label_set:
            yield {token_label}
        else:
            yield label_set


def filter_distribution(constraints, threshold):
    """
    Filter the POS associated to a word according to the method used
    in [Täckström et al, 2013]
    """
    new_constraints = {}
    for key, c in constraints.items():
        n = sum(c.values())
        filtered_tags = {k for k, v in c.items() if v / n > threshold}
        # Do not add empty constraints
        if filtered_tags:
            new_constraints[key] = filtered_tags

    return new_constraints


# !!!Fix: should be splitted in two subfunctions
def filter_constraints(constraints, threshold):
    """
    Filter `constraints` set by applying a threshold, either:

    - Nothing if `threshold` is None
    - On the distribution, if `threshold` < 1
    - Keeping only the `threshold` most frequent POS in token constraints
    otherwise

    """

    if threshold is None:
        return {w: set(c) for w, c in constraints.items()}

    assert threshold >= 0

    if threshold < 1:
        logging.info("Using Täckström's filtering method (threshold: {})"
                     "".format(threshold))
        return filter_distribution(constraints, threshold)

    else:
        logging.info("Using most frequent filtering method (max: {})"
                     "".format(threshold))
        threshold = int(threshold)

    # !!!Fix: fix 'bug' here. Current implementation may return more than
    # `threshold` pos when frequencies are equal. (Should this be fixed
    # or is this ok and better than random (or heuristic) choice ?.
    new_constraints = {}
    for key, c in constraints.items():
        values = {v for v in c.values()}
        limit = min(threshold - 1, len(values) - 1)
        max_values = sorted(values, reverse=True)[limit]

        new_constraints[key] = {k for k, v in c.items() if v >= max_values}

    return new_constraints


def filter_constraints_from_file(filename, output, threshold, listener=None):
    constraints = pickle.load(open(filename, "rb"))
    res = filter_constraints(constraints, threshold)

    with open(output, "wb") as ofile:
        pickle.dump(res, ofile, pickle.HIGHEST_PROTOCOL)


# !!!Fix: Weird argument order
def label_example(example,
                  wiki_constraints,
                  use_token,
                  al_constraints,
                  priority_constraints,
                  merge="intersection"):
    """Label an example using:

    - As type constraints, `wiki_constraints` and `al_constraints` using
    the `merge` heuristic to combine them.

    - Token constraints, if `use_token`

    """

    words = example["tgt"]
    labels = apply_type_constraints(sentence=words,
                                    wiki_constraints=wiki_constraints,
                                    al_constraints=al_constraints,
                                    priority_constraints=priority_constraints,
                                    merge=merge)

    if use_token:
        labels = apply_token_constraints(labels, example)

    return words, labels


# !!!Fix: Weird argument order
def label_corpus(wiki_constraints,
                 use_token,
                 al_constraints,
                 priority_constraints,
                 corpus_fh,
                 max_ex,
                 merge="intersection",
                 listener=None):
    """
    Proxy method to apply type and token constraints (in this order)

    Parameters
    ----------
    - corpus_fh, a filehandler
        the pickled file (open in "rb" mode) containing the aligned corpus.
    - wiki_constraints, a mapping
    - al_constraints, a mapping
    - max_ex, int
        the number of examples to consider
    - use_token, boolean
        whether to apply the token constraints or not
     """

    if listener is not None:
        data = listener.iter(pickle_reader(corpus_fh, max_ex))
    else:
        data = pickle_reader(corpus_fh, max_ex)

    for example in data:

        yield label_example(example, wiki_constraints, use_token,
                            al_constraints, priority_constraints, merge)
