from collections import Counter

from project_labels import apply_type_constraints, filter_constraints,\
    apply_token_constraints, label_example


from universal_pos import tagset as universal_tagset


def test_priority():
    """
    Apply priority constraints for "strange" punctuation
    """
    import regex
    import string

    wiki_constraints = {s: {"."} for s in string.punctuation}
    al_constraints = wiki_constraints

    res = apply_type_constraints(". ! ' ".split(),
                                 wiki_constraints,
                                 al_constraints,
                                 {})
    assert res == [{"."}] * 3

    res = apply_type_constraints(". ! ' « ".split(),
                                 wiki_constraints,
                                 al_constraints,
                                 {})
    assert res == [{"."}] * 3 + [set()]

    res = apply_type_constraints(". ! ' « ".split(),
                                 wiki_constraints,
                                 al_constraints,
                                 {regex.compile("«"): {"."}})
    assert res == [{"."}] * 4


def test_type_constraint_simple():
    """
    Apply type constraints in a simple case
    """
    al_constraints = {"maman": {"A"},
                      "et": {"C"},
                      "pouet": {"A"},
                      "le": {"E"}}

    wiki_constraints = {"maman": {"A", "B"},
                        "et": {"D"},
                        "bébé": {"F"}}

    # the example is chosen so as to include a word that do not appear in any
    # type or token constraints
    res = apply_type_constraints("la maman et le bébé".split(),
                                 wiki_constraints, al_constraints, set())

    expected_res = [set(), {"A"}, {"D"}, {"E"}, {"F"}]

    assert res == expected_res


def test_type_constraints_simple_union():
    al_constraints = {"maman": {"A"},
                      "et": {"C"},
                      "pouet": {"A"},
                      "le": {"E"}}

    wiki_constraints = {"maman": {"A", "B"},
                        "et": {"D"},
                        "bébé": {"F"}}

    # the example is chosen so as to include a word that do not appear in any
    # type or token constraints
    res = apply_type_constraints("la maman et le bébé".split(),
                                 wiki_constraints,
                                 al_constraints,
                                 set(),
                                 "union")

    expected_res = [set(), {"A", "B"}, {"D", "C"}, {"E"}, {"F"}]

    assert res == expected_res


def test_type_constraints_simple_union_with_priority():
    import regex
    al_constraints = {"maman": {"A"},
                      "et": {"C"},
                      "pouet": {"A"},
                      "le": {"E"}}

    wiki_constraints = {"maman": {"A", "B"},
                        "et": {"D"},
                        "bébé": {"F"}}

    # the example is chosen so as to include a word that do not appear in any
    # type or token constraints
    res = apply_type_constraints("la maman et le bébé".split(),
                                 wiki_constraints,
                                 al_constraints,
                                 {regex.compile("le"): {"AA"}},
                                 "union")

    expected_res = [set(), {"A", "B"}, {"D", "C"}, {"AA"}, {"F"}]

    assert res == expected_res


def test_empty_type():
    """
    What happens when there are no wiki constraints and only alignment
    constraints?
    """
    wiki_const = {}
    al_const = {"guillaume": {"B", "C"}}

    res = apply_type_constraints(["Guillaume"], wiki_const, al_const, set())
    expected_res = [{"B", "C"}]

    assert res == expected_res
    
    wiki_const, al_const = al_const, wiki_const

    res = apply_type_constraints(["Guillaume"], wiki_const, al_const, set())
    expected_res = [{"B", "C"}]

    assert res == expected_res
    

def test_real_example():
    """
    Test on a real example
    """

    wiki_constraints = {"με": {'PRON', 'ADP'}, ".": {"."}}

    al_constraints = {"με": universal_tagset,
                      "pouet": set(),
                      "guillaume": {'VERB', 'NOUN'}}

    res = apply_type_constraints("Με λένε Guillaume .".split(),
                                 wiki_constraints, al_constraints, set())

    expected_res = [{"PRON", "ADP"},
                    set(),
                    {"VERB", "NOUN"},
                    {"."}
                    ]

    assert res == expected_res


def test_count():
    """
    Test methods when alignment constraint include count information
    """
    al_const = {"a": Counter({"A": 2, "B": 3}),
                "c": Counter(),
                "b": Counter({"B": 3, "A": 1})}
    wiki_const = {"a": {"B", "A"}, "b": {"A"}}

    res = apply_type_constraints("a b".split(), wiki_const, al_const, set())
    expected_res = [{"B", "A"}, {"A"}]

    assert res == expected_res


def test_count_union():
    """
    Test methods when alignment constraint include count information
    """
    al_const = {"a": Counter({"A": 2, "B": 3}),
                "c": Counter(),
                "b": Counter({"B": 3, "CD": 1})}
    wiki_const = {"a": {"B", "A"}, "b": {"A"}}

    res = apply_type_constraints("a b c".split(),
                                 wiki_const,
                                 al_const,
                                 set(), "union")
    expected_res = [{"B", "A"}, {"A", "B", "CD"}, set()]

    assert res == expected_res


def test_further_filter():
    """
    Test method when alignment constraints can be used to filter wiki
    constraints
    """
    wiki_constraints = {"a": {"A", "B"},
                        "b": {"B", "A"}}
    al_constraints = {"a": {"A", "B"},
                      "b": {"A"},
                      "pouet": None}

    res = apply_type_constraints("a b".split(),
                                 wiki_constraints, al_constraints, set())
    expected_res = [{"A", "B"}, {"A"}]

    assert res == expected_res


def test_no_intersection():
    """
    Build the ambiguous label set when wiki and alignment constraints have no
    intersection
    """
    wiki_constraints = {"a": {"A", "B", "C"}}
    al_constraints = {"a": "D", "pouet": "poeut"}

    res = apply_type_constraints("a", wiki_constraints, al_constraints, set())
    expected_res = [{"A", "B", "C"}]

    assert res == expected_res


def test_apply_token():
    """
    Apply token constraints
    """

    src = ['gold', 'prices', 'even', 'hit', 'a', 'record-high',
           '$1,300', 'recently', '.']
    tgt = ['los', 'precios', 'de', 'el', 'oro', 'incluso', 'alcanzaron',
           'recientemente', 'un', 'récord', 'de', '1.300', 'dólares', '.']
    al = {0: [4], 1: [1], 2: [5], 3: [6], 4: [8], 5: [9], 6: [11], 7: [
        7], 8: [13]}
    src_pos = ['NOUN', 'NOUN', 'ADV', 'VERB', 'DET', 'ADJ', 'NOUN', 'ADV', '.']
    tgt_pos = ['DET', 'NOUN', 'ADP', 'DET', 'NOUN', 'ADV', 'VERB', 'ADV',
               'DET', 'NOUN', 'ADP', 'ADJ', 'NOUN', '.']

    example = {'src': src,
               'tgt': tgt,
               'al': al,
               'src_pos': src_pos,
               'tgt_pos': tgt_pos}

    labels = [{'PRON', 'DET'}, {'NOUN'}, {'NOUN', 'ADP'}, {'DET'},
              {'NOUN'}, {'ADV', 'VERB'}, {'VERB'}, {'ADV'}, {'DET'}, {'NOUN'},
              {'NOUN', 'ADP'}, {'NUM', 'NOUN'}, {'NOUN'}, {'.'}]

    expected_labels = [{'PRON', 'DET'}, {'NOUN'}, {'NOUN', 'ADP'}, {'DET'},
                       {'NOUN'}, {'ADV'}, {'VERB'}, {'ADV'}, {'DET'}, {'NOUN'},
                       {'NOUN', 'ADP'}, {'NOUN'}, {'NOUN'}, {'.'}]

    labels = apply_token_constraints(labels, example)
    print(labels)
    assert expected_labels == labels


def test_apply_only_token():
    """
    Apply only token constraints
    """

    src = ['gold', 'prices', 'even', 'hit', 'a', 'record-high',
           '$1,300', 'recently', '.']
    tgt = ['los', 'precios', 'de', 'el', 'oro', 'incluso', 'alcanzaron',
           'recientemente', 'un', 'récord', 'de', '1.300', 'dólares', '.']
    al = {0: [4], 1: [1], 2: [5], 3: [6], 4: [8], 5: [9], 6: [11], 7: [
        7], 8: [13]}
    src_pos = ['NOUN', 'NOUN', 'ADV', 'VERB', 'DET', 'ADJ', 'NOUN', 'ADV', '.']
    tgt_pos = ['DET', 'NOUN', 'ADP', 'DET', 'NOUN', 'ADV', 'VERB', 'ADV',
               'DET', 'NOUN', 'ADP', 'ADJ', 'NOUN', '.']

    example = {'src': src,
               'tgt': tgt,
               'al': al,
               'src_pos': src_pos,
               'tgt_pos': tgt_pos}

    expected_labels = [set(), {'NOUN'}, set(), set(), {'NOUN'},
                       {'ADV'}, {'VERB'}, {'ADV'}, {'DET'}, {'ADJ'}, set(), {'NOUN'},
                       set(), {'.'}]

    words, labels = label_example(example,
                                  set(),
                                  True,
                                  set(),
                                  set())
    assert words == example["tgt"]
    assert expected_labels == list(labels)


# ===================
# Test filter method
# ===================


def test_single_filter():
    """
    Filter a dictionnary with a single elements
    """
    constraints = {"A": Counter({"a": 3, "b": 4, "c": 5})}
    filtered_constraints = {"A": {"b", "c"}}

    assert filtered_constraints == filter_constraints(constraints, 2)


def test_filter_same_count():
    """
    Filter a dictionnary when several elements have the same count
    """

    constraints = {"A": Counter({"a": 3, "b": 4, "c": 5, "d": 4})}
    filtered_constraints = {"A": {"b", "c", "d"}}

    assert filtered_constraints == filter_constraints(constraints, 2)
    assert {"A": {"c"}} == filter_constraints(constraints, 1)


def test_filter_several():
    """
    Filter a dicitonnary with several keys
    """

    constraints = {"A": Counter({"a": 3, "b": 4, "c": 5, "d": 4}),
                   "B": Counter({"a": 5, "e": 3}),
                   "C": Counter({"e": 1, "f": 8, "d": 5, "g": 6})}

    assert {'A': {'c', 'b', 'd'}, 'B': {'a', 'e'}, 'C': {'g', 'f'}
            } == filter_constraints(constraints, 2)
    assert {'A': {'c'}, 'B': {'a'}, 'C': {'f'}} == filter_constraints(
        constraints, 1)


def test_filter_too_small():
    """
    Filter token constraints when there are not enough element
    """

    constraints = {"A": Counter({"a": 4})}
    expected_constraints = {"A": {"a"}}

    assert expected_constraints == filter_constraints(constraints, 3)


def test_filter_threshold():
    """
    Filter the label sets using Tackström method
    """
    token_constraints = {"A": {"VERB": 100, "NOUN": 1},
                         "B": {"VERB": 1},
                         "C": {"NOUN": 10, "VERB": 10}}

    expected_res = {"A": {"VERB"},
                    "B": {"VERB"},
                    "C": {"NOUN", "VERB"}}

    assert filter_constraints(token_constraints, .2) == expected_res

    expected_res = {"A": {"VERB", "NOUN"},
                    "B": {"VERB"},
                    "C": {"NOUN", "VERB"}}

    assert filter_constraints(token_constraints, 2) == expected_res
