import pickle

from functools import partial
from collections import namedtuple, defaultdict

from universal_pos import tagset as universal_tagset


def read_type_constraint(filehandler):
    """
    Read type constraints in the format used in [Li et al, 2012] paper (each
    line contains a word and one POS tag it can have separated by a
    tabulation); the same word can appear on several lines.
    """

    from itertools import groupby
    from operator import itemgetter

    const = (f.strip().split("\t") for f in filehandler)
    constraints = {k: {vv for _, vv in v} for k, v in
                   groupby(const, key=itemgetter(0))}

    return constraints


def pickle_reader(filehandler, n_ex=None):
    """
    Read `n_ex` examples from a corpus pickled in `filehandler`.
    """

    if n_ex == -1:
        n_ex = None

    count = 1
    while True:

        try:
            line = pickle.load(filehandler)
        except EOFError:
            break

        yield line

        count += 1
        if n_ex is not None and count > n_ex:
            break


def prepare_test_data(corpus,
                      wiki_constraints,
                      alignment_constraints,
                      priority_constraints):
    """
    Compute the set of possible labels for each word of a corpus

    During inference, only a label can be predicted. This constraints are
    derived from the alignment and wiktionnary constraints.

    Parameters:
    -----------
    - corpus, XXX
      the dataset to consider
    - wiki_constraint, a mapping
      the type constraints extracted from Wiktionnary
    - alignment_constraint, a mapping
      the type constraints extracted from the alignments
    """

    from project_labels import apply_type_constraints

    ambiguous_label = partial(apply_type_constraints,
                              wiki_constraints=wiki_constraints,
                              al_constraints=alignment_constraints,
                              priority_constraints=priority_constraints)

    relabel_empty = lambda x: [ll if ll else universal_tagset for ll in x]
    return [(w, l, relabel_empty(ambiguous_label(w))) for w, l in corpus]



def read_priority_constraints(priority_const):

    import json
    import regex

    if priority_const is None:
        return set()

    priority_const = json.load(open(priority_const, "rt"))
    priority_const = {regex.compile(k): set(v)
                      for k, v in priority_const.items()}

    return priority_const


class EmptyToken:

    def __getattr__(self, name):
        return None


empty_token = EmptyToken()


def conll_reader(filehandler, header):
    """
    Read a file formatted as a table
    """
    import sys

    Token = namedtuple("Token", header)

    cur_obser = []
    for line in filehandler:
        line = line.strip().split()

        if not line:
            cur_obser = defaultdict(lambda: empty_token, enumerate(cur_obser))
            if cur_obser:
                yield cur_obser
            cur_obser = []
            continue

        try:
            cur_obser.append(Token(*line[:len(header)]))
        except Exception as e:
            print("ERREUR")
            print(e)
            print(line)
            sys.exit(1)

    if len(cur_obser) != 0:
        cur_obser = defaultdict(lambda: empty_token, enumerate(cur_obser))
        yield cur_obser


def load_conll(filehandler):
    dataset = conll_reader(filehandler,
                           "ID FORM A CPOS FPOS B C D E F".split())
    dataset = [[s[idx] for idx in sorted(s)] for s in dataset]
    dataset = [[(t.FORM, t.CPOS) for t in s] for s in dataset]
    return [list(zip(*t)) for t in dataset]


def read_test(test_set):

    test_set = open(test_set, "rt")
    test_set = conll_reader(test_set,
                            "ID FORM A CPOS FPOS B C D E F".split())

    test_set = [[sentence[pos] for pos in sorted(sentence)]
                for sentence in test_set]
    test_set = [[(t.FORM, t.CPOS) for t in sentence]
                for sentence in test_set]

    return [list(zip(*t)) for t in test_set]
