"""
High-level API for training POS tagger (either supervised or weakly-supervised)
"""
import json
import pickle
import logging

from sequence_labeling.evaluation import hamming_loss, amb_hamming_loss
from sequence_labeling.laso_tagger import POSTagger, AveragedPerceptron, WeakilySupervisedPerceptron, SkipPerceptron
from weakly_io import pickle_reader, read_priority_constraints, read_type_constraint
from weakly_io import load_conll
logging.basicConfig(level=logging.INFO)

from exp_weakly import make_feature_template, make_type_constraints, get_type_constraints


def project_labels(output_filename,
                   corpus,
                   use_token,
                   wiktionary_fn,
                   al_constraints_fn,
                   priority_constraints_fn,
                   merge,
                   listener,
                   max_ex=-1):
    """
    Main method to label a dataset by projecting labels accross
    alignment links
    """

    import pickle

    from project_labels import label_corpus

    ofile = open(output_filename, "wb")

    wiki_cons = read_type_constraint(open(wiktionary_fn, "rt")) if wiktionary_fn is not None else set()
    al_constraints = pickle.load(open(al_constraints_fn, "rb")) if al_constraints_fn is not None else set()
    priority_cons = read_priority_constraints(priority_constraints_fn)

    if merge is None:
        assert wiki_cons == set() or al_constraints == set()
        merge = "union"

    for ex in label_corpus(wiki_constraints=wiki_cons,
                           use_token=use_token,
                           alignment_constraints=al_constraints,
                           priority_constraints=priority_cons,
                           corpus_fh=open(corpus, "rb"),
                           merge=merge,
                           max_ex=max_ex,
                           listener=listener):
        pickle.dump(ex, ofile, pickle.HIGHEST_PROTOCOL)

    return output_filename


def train_supervised_tagger(lang, train_set, features, output_fn,
                            constraints=None, listener=None):
    """
    Main function to train a supervised POS tagger:

    Parameters
    ----------
    - lang: a string
          a description of the language the tagger is trained
          for. Mainly used for logging purpose
    - train_set: a list of tuples (words, gold_labels)
          the corpus to use for training
    - output_fn, a string
          the name of the file the model has to be dumped to
    - constraints
    """
    logging.info("train supervised tagger for {}".format(lang))

    features = get_features(lang, features)
    train_set = load_conll(open(train_set, "rt"))

    n_iter = 5

    if constraints is not None:
        constraints = make_type_constraints(lang, constraints)

    tagger = POSTagger(features, True, AveragedPerceptron)
    tagger.train(train_set, n_iter, possible_labels=constraints, listener=listener)

    pickle.dump(tagger, open(output_fn, "wb"), pickle.HIGHEST_PROTOCOL)

    return output_fn


def get_features(lang, features):
    from exp_weakly import history_size

    if features == "basic_features":
        return make_feature_template(lang, history_size, False)
    elif features == "constraints_features":
        return make_feature_template(lang,
                                     history_size,
                                     make_type_constraints(lang, "full"))
    else:
        raise Exception


def train_weakly_tagger(lang, train_set_fn, features, output_fn,
                        constraints=None, skip=False, listener=None, n_ex=None):

    from exp_weakly import n_ex as default_n_ex, weakly_n_iter as n_iter, make_type_constraints

    logging.info("train weakly tagger for {}".format(lang))

    if n_ex is None:
        n_ex = default_n_ex

    listener.msg("n_iter = {:,}".format(n_iter))
    listener.msg("n_ex = {:,}".format(n_ex))

    train_set = open(train_set_fn, "rb")
    train_set = list(pickle_reader(train_set, n_ex))
    features = get_features(lang, features)

    if constraints is not None:
        constraints = make_type_constraints(lang, constraints)

    if not skip:
        logging.info("Using a weakly supervised perceptron")
        tagger = POSTagger(features, True, WeakilySupervisedPerceptron)
    else:
        logging.info("Using a skip perceptron")
        tagger = POSTagger(features, True, SkipPerceptron)

    n_iter = 1
    tagger.train(train_set, n_iter, possible_labels=constraints, listener=listener)
    pickle.dump(tagger, open(output_fn, "wb"), pickle.HIGHEST_PROTOCOL)

    return output_fn


def test_tagger(tagger_fn, test_set_fn, lang, type_constraints=None, metric="hamming"):
    """
    Parameters
    ----------

    - tagger_fn, a string
      the name of file the tagger has been serialized to
    - test_set_fn, a string
      the name fo the file containing the test set
    - lang, a string
      the language used in the experiment (used to load extra
      resources)
    - type_constraints, a string
      if `None`, type constraints are not use to constrain the label
      set during predictions; 
     
      It can take one of the following value: 
    """

    test_set = load_conll(open(test_set_fn, "rt"))
    sent, gold = zip(*test_set)
    model = pickle.load(open(tagger_fn, "rb"))

    if type_constraints is not None:
        type_constraints = get_type_constraints(lang, type_constraints)
        possible_labels = [list(map(type_constraints, s)) for s in sent]
        predicted = [model.tag(s, p) for s, p in zip(sent, possible_labels)]
    else:
        predicted = [model.tag(s) for s in sent]

    if metric == "hamming":
        return hamming_loss(zip(predicted, gold))
    elif metric == "amb_hamming":
        type_constraints = make_type_constraints(lang, "full")
        possible_labels = [list(map(type_constraints, s)) for s in sent]
        return amb_hamming_loss(predicted, gold, possible_labels)
    else:
        raise Exception
